import os.path #To get a file name to write to.
from typing import cast, List
from UM.Application import Application #To find the scene to get the current g-code to write.
from UM.FileHandler.WriteFileJob import WriteFileJob #To serialise nodes to text.
from UM.Logger import Logger
from UM.OutputDevice.OutputDevice import OutputDevice #An interface to implement.
from UM.OutputDevice.OutputDevicePlugin import OutputDevicePlugin #The class we need to extend.
from UM.PluginRegistry import PluginRegistry
from UM.Mesh.MeshWriter import MeshWriter
import requests
import io

# IP ADDRESS --------------
IPADDRESS = "192.168.0.39"
# -------------------------


class SendToSnapmaker2Plugin(OutputDevicePlugin): 
    def start(self):
        self.getOutputDeviceManager().addOutputDevice(SendToSnapmaker2())
    def stop(self):
        self.getOutputDeviceManager().removeOutputDevice("Snapmaker2")

class SendToSnapmaker2(OutputDevice):
    def __init__(self):
        super().__init__("Snapmaker2") 

        self.setName("Snapmaker2.0")
        self.setShortDescription("Send to Snapmaker2.0")
        self.setDescription("Send .gcode to Snapmaker2.0")
        self.setIconName("save")

        self.url = "http://{}:8080/api/v1".format(IPADDRESS)
        self.token = ""

    def requestWrite(self, nodes, file_name = None, limit_mimetypes = None, file_handler = None, **kwargs):
        file_writer = cast(MeshWriter, PluginRegistry.getInstance().getPluginObject("GCodeWriter"))
        output_file_name = os.path.expanduser("~/output.txt")
        file_stream = io.StringIO()
        
        job = WriteFileJob(file_writer, file_stream, nodes, 1)
        job.progress.connect(self._onProgress)
        job.finished.connect(self._onFinished)
        job.start()

    def _onProgress(self, job, progress):
        Logger.log("d", "Saving file... {progress}%".format(progress = progress))

    def _onFinished(self, job):
        # attempt to fetch token
        try: 
            self.token = self._get_token(self.url)
            if self.token == "":
                raise Exception
        except:
            Logger.log("e","Couldn't connect to printer")
            job.getStream().close()
            return

        Logger.log("d", "Token: {}".format(self.token))

        # wait for touchscreen press
        status = self._get_status(self.url, self.token)
        while status.status_code == 204:
            status = self._get_status(self.url, self.token)

        # attempt upload
        if status.status_code == 200:
            try:
                file = job.getStream()
                Logger.log("d","Attempting upload...")
                r = self._post_file(self.url, self.token, file)
                if r.status_code != 200:
                    raise Exception
            except:
                Logger.log("d", r.content)
                job.getStream().close()
                return
        else:
            Logger.log("e","Upload rejected")
            job.getStream().close()
            return

        job.getStream().close()
        Logger.log("d", "Done saving file!")

    def _get_token(self, url):
        """ requests session token """
        r = requests.post(url+"/connect")
        if r.status_code == 200:
            return r.json()["token"]
        else:
            return ""

    def _get_status(self, url, token):
        """ request printer status """
        payload = {"token": token}
        r = requests.get(url+"/status", params=payload)
        return r

    def _post_file(self, url, token, file):
        """ upload file to printer """
        payload = {"token": token}
        file_payload = {"file": file.getvalue()}
        r = requests.post(url+"/upload", params=payload, files=file_payload)
        return r